package tsp.csc4509.blockfile;


import static tsp.csc4509.common.Log.GEN;


import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;





public class BlockFile implements AutoCloseable {
	/**
	 * répertoire où sont placés les fichiers à lire ou écrire.
	 */
	public static final String DATAREP = "data";
	/**
	 * répertoire où sont écrits les fichiers créés par les copies ou les imports. 
	 */
	public static final String COPY = DATAREP + "/" + "copy";
	/**
	 * répertoire où sont placés les fichiers sources.
	 */
	public static final String SOURCE = DATAREP + "/" + "source";
	/**
	 * répertoire où sont placés les fichiers info.
	 */
	public static final String INFO = DATAREP + "/" + "info";
	/**
	 * extension des  fichiers info.
	 */
	public static final String INFOEXT = ".info";

	/**
	 * taille des blocs pour le générateur des fichiers info.
	 */
	public static final int BLOCKSIZE = 10000;

	/**
	 * Nom du fichier à copier ou importer.
	 */
	private String fileName;


	/**
	 * Taille du fichier (en octets).
	 */
	private long size;

	/**
	 * Ensemble des blocs qui ne sont pas encore copiés ou importés. 
	 */
	private Set<Block> blocksSet;


	/**
	 * Référence sur le fichier ouvert en écriture en random access.
	 */
	private RandomAccessFile rout;

	/**
	 * Canal relié au fichier rout.
	 */
	private FileChannel fcout;



	/**
	 * Constructeur qui:
	 *  - ouvre dans le répertoire COPY le fichier en écriture;
	 *  - ouvre le canal relié à ce fichier;
	 *  - charge les informations de l'ensemble blokcsSet.
	 * @param fileName
	 *    nom du fichier à importer (sans chemin: il sera placé dans le répertoire COPY).
	 * @throws IOException
	 *     toutes les exceptions qui peuvent se produire lors de la manipulation des fichiers.
	 */
	public BlockFile(final String fileName) throws  IOException {
		this.fileName = fileName;
		rout = new RandomAccessFile(COPY + "/" + fileName, "rw");
		fcout = rout.getChannel();
		this.createBlocksSet(fcout);
	}

	@Override
	public void close() throws IOException {
		rout.close();
		fcout.close();
	}

	/**
	 * Retire un bloc de l'ensemble des blocs à importer (blocksSet).
	 * @return
	 *       null si l'ensemble est vide, et sinon le bloc pris dans l'ensemble.
	 */
	public Block getABlock() {
		Iterator<Block> iterator = blocksSet.iterator();
		Block block = null;
		if (iterator.hasNext()) {
			block = iterator.next();
			iterator.remove();
		}
		return block;
	}

	/**
	 * Ajoute un bloc dans l'ensemble des blocs à importer (blocksSet).
	 * @param block
	 */
	public void addBlock(final Block block) {
		blocksSet.add(block);
	}

	/**
	 * Crée et initialise l'ensemble blockSet. (Méthode privée appelée par le constructeur).
	 * Cette méthode utilise un fichier d'informations sur le fichier à importer ou à copier.
	 * Le fichier d'informations doit se trouver dans le répertoire INFOREP et porte le
	 * nom du fichier à importer suffixé de l'extension INFOEXT.
	 * Format du fichiers d'information:
	 *       . un long contenant la taille totale du fichier;
	 *       . un int contenant la taille des blocs du fichier.
	 * @param fcout
	 *        canal lié au fichier ou seront écrits les blocs lors de leur réception.
	 *             
	 * @throws IOException
	 *     toutes les exceptions qui peuvent se produire lors de la manipulation des fichiers.
	 */
	private void createBlocksSet(final FileChannel fcout) throws  IOException {
		Block bloc;
		int blockSize;

		try (FileInputStream fin = new FileInputStream(INFO +  "/" + fileName + INFOEXT);
				FileChannel fc = fin.getChannel();) {
			ByteBuffer fileSizeBuffer = ByteBuffer.allocate(Long.BYTES);
			ByteBuffer blockSizeBuffer = ByteBuffer.allocate(Integer.BYTES);

			// lecture de la taille du fichier.
			fc.read(fileSizeBuffer);
			fileSizeBuffer.flip();
			size = fileSizeBuffer.getLong();
			// lecture de la taille des blocs.
			fc.read(blockSizeBuffer);
			blockSizeBuffer.flip();
			blockSize = blockSizeBuffer.getInt();
			int nbBlocs = (int) (size / blockSize);
			if (size % blockSize != 0) {
				// il y a un reliquat à gérer
				nbBlocs++;
			}
			long offset = 0L;
			blocksSet = new HashSet<Block>();
			// création de tous les block.
			for (int num = 0; num < nbBlocs; num++) {
				if (num == nbBlocs - 1) {
					// size of last block 
					blockSize = (int) (size - blockSize * num); 
				}
				bloc = new Block(offset, blockSize, fcout);
				blocksSet.add(bloc);
				GEN.trace("Taille = " + bloc.getSize() + ", offset " + bloc.getOffset());
				offset += blockSize;
			}

			GEN.trace("Nombre de bloc " + blocksSet.size());
		} 
	}

	/**
	 * getter sur le nom du fichier à importer ou à copier.
	 * @return
	 *        le nom du fichier à importer ou à copier.
	 */
	public String getFileName() {
		return fileName;
	}


	/**
	 * Indique s'il reste des blocs non traités.
	 * @return
	 *       true s'il reste des blocs non traités dans l'ensemble blockSet.
	 */
	public boolean hasRemaining() {
		return blocksSet.size() > 0;
	}

	/**
	 * Crée le fichier contenant les infos sur le découpage du fichier.
	 * Le fichier d'informations doit se trouver dans le répertoire INFOREP et porte le
	 * nom du fichier à importer suffixé de l'extension INFOEXT.
	 * Format du fichiers d'information:
	 *       . un long contenant la taille totale du fichier;
	 *       . un int contenant la taille des blocs du fichier.
	 * @param fileName
	 *     nom du fichier (sans aucun chemin avant). Le fichier doit se trouver dans le 
	 *     répertoire SOURCE.
	 * @throws IOException
	 *     toutes les exceptions qui peuvent se produire lors de la manipulation des fichiers.
	 */
	public static void createFile(final String fileName) throws IOException {
		

		GEN.trace("creation du fichier Info pour " + fileName);
		try (RandomAccessFile rin = new RandomAccessFile(SOURCE + "/" + fileName, "r");
				FileOutputStream fout = new FileOutputStream(INFO + "/" + fileName + INFOEXT);
				FileChannel fcin = rin.getChannel();
				FileChannel fcout = fout.getChannel();) {
			long size = rin.length();
			ByteBuffer fileSizeBuffer = ByteBuffer.allocate(Long.BYTES);
			ByteBuffer blockSizeBuffer = ByteBuffer.allocate(Integer.BYTES);
			fileSizeBuffer.putLong(size);
			fileSizeBuffer.flip();
			// taille du fichier à découper au début du fichier des digests.
			fcout.write(fileSizeBuffer);
			blockSizeBuffer.putInt(BLOCKSIZE);
			blockSizeBuffer.flip();
			// taille des blocs à découper mis en deuxième position du fichier des digests.
			fcout.write(blockSizeBuffer);
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}

	}

}


