package tsp.csc4509.tpnote;

import static tsp.csc4509.common.Log.GEN;
import java.io.IOException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import tsp.csc4509.blockfile.Block;
import tsp.csc4509.blockfile.BlockFile;
import tsp.csc4509.tcp.TcpSocketSynchrone;

/**
 * Classe du client.
 * @author
 *
 */
public class Client {

	/**
	 * nombre de connexions simultanées ouvertes par le client.
	 */
	private static final int NBCONNECTIONS = 5;


	/**
	 * Fichier divisé en blocs attendu par le client.
	 */
	private BlockFile blockFile;

	/**
	 * map contenant l'ensemble des blocs actuellement en téléchargement (clef: le canal connecté au serveur).
	 */
	private Map<SocketChannel, Block> serverMap;

	/**
	 * selector utilisé par le client pour gérer son mode asynchrone. 
	 */
	private Selector selector;


	
	/**
	 * Contructeur du client qui.
	 *    -> initialise file avec un File gérant le fichier fileName; 
	 *    -> initialise serverMap avec une HashMap;
	 *    -> ouvre le selector;
	 * @param fileName
	 *        nom du fichier à importer.
	 * @throws IOException
	 *        toutes les exceptions d'entrées sorties qui peuvent se produire.
	 */
	public Client(final String fileName) throws IOException {
		super();
		GEN.info("Construction du client pour " + fileName);
		
		blockFile = new BlockFile(fileName);
		serverMap = new HashMap<SocketChannel, Block>();
		selector = Selector.open();
	}

	/**
	 * Connecte le client au serveur, et lui envoie une requête.
	 * @param serverInfo
	 *         coordonnées (hostname et port) du serveur.
	 * @param request
	 *         requête à envoyer.
	 * @return
	 *         le canal de la connexio reliée au serveur.
	 * @throws IOException
	 *         toutes les exceptions d'entrées/sorties qui peuvent se produire.
	 */
	public SocketChannel connectAndSendRequest(final ServerInfo serverInfo, final Request request) throws IOException { // devrait être private. Rendue public pour la classe de test
		GEN.info("{}", () ->  "Client::connectAndSendRequest sur " + serverInfo.getHostname() + ":" + serverInfo.getPort());
		TcpSocketSynchrone connection = new TcpSocketSynchrone(serverInfo.getHostname(), serverInfo.getPort());
		connection.sendObject(request);
		return connection.getRwChan();
	}


	/**
	 * Méthode qui prend en charge un nouveau bloc du fichier à importer.
	 * Cette méthode doit:
	 *        1. obtenir un bloc du fichier qui n'est pas encore téléchargé (s'il en reste);
	 *        2. créer un nouvelle requête;
	 *        3. se connecter à un serveur en envoyer cette requête;
	 *        4. passer en mode asynchrone, inscrire le canal relié au serveur dans le selector
	 *           et ajouter ce bloc dans la map qui gère les blocs en téléchargement.
	 * @throws IOException
	 *           toutes les exceptions d'entrées/sorties qui peuvent se produire.
	 */
	public void manageNextBlock() throws IOException  { // devrait être private. Rendue public pour la classe de test
		GEN.info("Client::manageNextBlock() débuté");
		Block block = blockFile.getABlock();
		if (block == null) {
			GEN.info("Client::manageNextBlock() tous les blocs déjà traités");
			// il ne reste plus de bloc vide à gérer.
			return;
		}
		Request request = new Request(blockFile, block);
		
		SocketChannel rwChan = null;
		do {
			ServerInfo serverInfo = ServerInfo.getRandomServer();
			GEN.info("Client::manageNextBlock() Server:" + serverInfo);
			try {
				rwChan = connectAndSendRequest(serverInfo, request);
			} catch (IOException e) {
				GEN.info("Client::manageNextBlock() récupération d'une erreur de connexion");
				// rien à faire, la boucle va retenter avec un autre serveur.
			}
		} while (rwChan == null);

		// une fois la requête envoyée on passe en mode asynchrone pour la réception de réponse.
		rwChan.configureBlocking(false);
		rwChan.register(selector, SelectionKey.OP_READ);
		serverMap.put(rwChan, block);
		GEN.info("Client::manageNextBlock() nouveau bloc pris en charge");
	}


	/**
	 * Méthode qui importe le fichier attendu par ce client.
	 * Cette méthode doit:
	 *      1. Mettre en place la gestion des NBCONNECTIONS premiers blocs;
	 *      2. Et tant qu'il reste un bloc à importer:
	 *        2a. Gérer en simultané la réception des données de tous les blocs gérés;
	 *        2b. Fermer tout bloc complété, et prendre en charge un nouveau bloc.
	 * Si le client perd la connexion avec un serveur avant que le bloc qu'il
	 * importe ne soit complet, il faut alors replacer ce blocs dans l'ensemble
	 * du fichier (pour qu'il puisse à nouveau être gérer plus tard), et prendre
	 * en charge un nouveau bloc.
	 *        
	 * @throws IOException
	 *           toutes les exceptions d'entrées/sorties qui peuvent se produire.
	 */
	public void importFile() throws IOException {

		// création des premieres connexions.
		for (int nb = 0; nb < NBCONNECTIONS; nb++) {
			manageNextBlock();
		}


		
		while (!serverMap.isEmpty() || (blockFile.hasRemaining())) {	
			
			selector.select();
			Set<SelectionKey> readyKeys = selector.selectedKeys();

			for (SelectionKey curKey : readyKeys) {
				SocketChannel rwChan = (SocketChannel) curKey.channel();
				Block block = serverMap.get(rwChan);
				if (block.fill(rwChan) == -1) {
					GEN.info("Client::importFile() ******************** connexion perdue");
					// connexion fermée avant remplissage du bloc
					serverMap.remove(rwChan);
					rwChan.close();
					// on remet le bloc dans l'ensemble des blocs à importer.
					block.close(); // il faut abandonner le buffer de ce bloc
					blockFile.addBlock(block);
					// on  prend en charge un nouveau bloc.
					manageNextBlock();
				} else {
					if (!block.hasRemaining()) {
						GEN.info("Client::importFile() un bloc complété");
						// le block est complet.
						block.close();
						serverMap.remove(rwChan);
						rwChan.close();
						manageNextBlock();
					}
				}
			}
			readyKeys.clear();
		}
	}


	// ***************** getters sur les attributs ajoutés à la classe pour l'écriture des tests
	// ils ne doivent pas être utilisés dans votre code. Seule la classe Client doit manipuler ces attributs.
	/**
	 * getter sur blockFile.
	 * @return
	 *     l'attribut blockFile.
	 */
	public BlockFile getBlockFile() {
		return blockFile;
	}

	/**
	 * getter sur serverMap.
	 * @return
	 *     l'attribut serverMap.
	 */
	public Map<SocketChannel, Block> getServerMap() {
		return serverMap;
	}

	/**
	 * getter sur selector.
	 * @return
	 *     l'attribut selector.
	 */
	public Selector getSelector() {
		return selector;
	}
	
}
